"""Tests for scale_to_zero_to_one.py."""
import unittest

import numpy as np
import pandas as pd
import xarray as xr

from satip.scale_to_zero_to_one import ScaleToZeroToOne, is_dataset_clean


class TestScaleToZeroToOne(unittest.TestCase):
    """Test class for methods of class scale_to_zero_to_one.ScaleToZeroToOne.

    We will set up a mock dataset and try the various methods in ScaleToZeroToOne,
    checking whether expected results manifest themselves.
    """

    def setUp(self) -> None:
        """Set up for all following tests, where a common dataarray is defined."""
        # Set dimensionality of the fake dataset:
        Nx, Ny, Nt = 2, 3, 10

        # Generate the fake four-dimensional data:
        data = np.zeros((Nx, Ny, Nt, 2))
        data[:, :, :, 0] = np.linspace(-10, 10, Nt, endpoint=True) + np.random.rand(Nx, Ny, Nt)
        data[:, :, :, 1] = np.linspace(10, -10, Nt, endpoint=True) + np.random.rand(Nx, Ny, Nt)

        # Set some random values in the middle to NaN:
        data[Nx // 2, Ny // 2, Nt // 2, :] = np.nan

        self.dataset = xr.DataArray(
            data=data,
            coords=dict(
                lon=(
                    ["x_geostationary", "y_geostationary"],
                    np.linspace(0, 1.0, Nx, endpoint=True).reshape((Nx, 1)) + np.zeros((Nx, Ny)),
                ),
                lat=(
                    ["x_geostationary", "y_geostationary"],
                    np.linspace(-1.0, 0.0, Ny, endpoint=True).reshape((1, Ny)) + np.zeros((Nx, Ny)),
                ),
                time=pd.date_range("2019-03-08", periods=Nt),
            ),
            dims=["x_geostationary", "y_geostationary", "time", "variable"],
            attrs=dict(
                description="Some randomly permutated lines in time and space.\
                     If you find meaning in this, please see a shrink."
            ),
        )

        return super().setUp()

    def test_fit(self):  # noqa: D102

        scaler = ScaleToZeroToOne(
            mins=np.asarray([-5, 0]),
            maxs=np.asarray([5, 20]),
            variable_order=["wrong_var_name_one", "wrong_var_name_two"],
        )
        scaler.fit(self.dataset, dims=("x_geostationary", "y_geostationary", "time"))

        # Test whether the min/max-values are logged:
        self.assertListEqual(
            scaler.mins.values.tolist(),
            self.dataset.min(("x_geostationary", "y_geostationary", "time"))
            .compute()
            .values.tolist(),
        )
        self.assertListEqual(
            scaler.maxs.values.tolist(),
            self.dataset.max(("x_geostationary", "y_geostationary", "time"))
            .compute()
            .values.tolist(),
        )

        # Test whether the initially wrong variable names are set correctly now:
        self.assertListEqual(scaler.variable_order.tolist(), [0, 1])

    def test_rescale(self):  # noqa: D102
        scaler = ScaleToZeroToOne().fit(
            self.dataset, dims=("x_geostationary", "y_geostationary", "time")
        )
        dataset = scaler.rescale(self.dataset)
        scaler.fit(dataset, dims=("x_geostationary", "y_geostationary", "time"))

        # Assert that all values are between zero and one:
        self.assertListEqual(scaler.mins.values.tolist(), [0, 0])
        self.assertListEqual(scaler.maxs.values.tolist(), [1, 1])

        # Are the NaN still in there?
        self.assertTrue(np.isnan(dataset).any())

    def test_compress_mask(self):  # noqa: D102
        # Generate a dataset and rescale it.
        # The result should be a dataset which still contains NaNs.
        scaler = ScaleToZeroToOne().fit(
            self.dataset, dims=("x_geostationary", "y_geostationary", "time")
        )
        dataset = scaler.rescale(self.dataset)

        # Now compress the dataset and then check if the NaN-values have been replaced with -1:
        dataset = scaler.compress_mask(dataset)

        self.assertTrue(dataset.min() == -1)
        self.assertFalse(np.isnan(dataset).any())

        # While we are at it, lets also test the is_dataset_clean-method:
        self.assertTrue(is_dataset_clean(dataset))



"""Tests for satip.download.py."""
import os
import unittest

import pandas as pd

from satip.download import (
    _determine_datetimes_to_download_files,
    _get_missing_datetimes_from_list_of_files,
    download_eumetsat_data,
)
from satip.utils import format_dt_str


class TestDownload(unittest.TestCase):
    """Test case for downloader tests."""

    def setUp(self) -> None:  # noqa
        return super().setUp()

    def test_download_eumetsat_data(self):  # noqa
        # Call the downloader on a very short chunk of data:
        self.assertIsNone(
            download_eumetsat_data(
                download_directory=str(os.getcwd() + "/storage/"),
                start_date=format_dt_str("2020-06-01 11:59:00"),
                end_date=format_dt_str("2020-06-01 12:02:00"),
                user_key=os.environ.get("EUMETSAT_USER_KEY"),
                user_secret=os.environ.get("EUMETSAT_USER_SECRET"),
                auth_filename=None,
                number_of_processes=2,
                product=["cloud", "rss"],
                enforce_full_days=False,
            )
        )

    def test_determine_datetime_to_download_files(self):
        """Tests correct day-wise-chunked lists.

        Given an empty directory, the function is supposed to return a complete
        list of dates, chunked into daily intervals.
        """
        datetimes = _determine_datetimes_to_download_files(
            ".",
            start_date=pd.to_datetime("2020-03-08 12:00"),
            end_date=pd.to_datetime("2020-03-10 09:00"),
            product_id="dummy",
        )
        print(datetimes)
        print(datetimes[0])
        print(type(datetimes[0][0]))
        self.assertEqual(datetimes[0][0], pd.to_datetime("2020-03-08 11:59:00"))
        self.assertEqual(datetimes[0][1], pd.to_datetime("2020-03-09 11:58:00"))
        self.assertEqual(datetimes[1][0], pd.to_datetime("2020-03-09 11:59:00"))
        self.assertEqual(datetimes[1][1], pd.to_datetime("2020-03-10 11:58:00"))

    def test_get_missing_datetimes_from_list_of_files(self):
        """Tests padding of datetimes if files present are missing data for given days."""
        # Assume we only have two files present, as something went very wrong
        # with the download.
        filenames = [
            "MSG3-SEVI-MSG15-0100-NA-20190308114036.810000000Z-NA.nat",
            "MSG3-SEVI-MSG15-0100-NA-20220308133711.810000000Z-NA.nat",
        ]

        # We then expect the function to pad this by adding missing timeranges
        # to fill up everything from the beginning to the first day to the timestamp
        # of the first day (case A), then everything between both files (case B)
        # and finally fill up the rest until the end of the second day (case C).
        expected_timestamps = [
            (pd.to_datetime("2019-03-08 00:00:00"), pd.to_datetime("2019-03-08 11:40:00")),
            (pd.to_datetime("2019-03-08 11:40:00"), pd.to_datetime("2022-03-08 13:37:00")),
            (pd.to_datetime("2022-03-08 13:37:00"), pd.to_datetime("2022-03-08 23:58:00")),
        ]

        # Generate the list of missing datetime interval boundaries:
        res = _get_missing_datetimes_from_list_of_files(filenames)

        for i, interval in enumerate(expected_timestamps):
            for b, boundary in enumerate(interval):
                # Note that the function returns datetime-objects, but we defined the expected
                # values as pd-datetime. Though their str-repr is different, they still pass
                # the equality-check when compared for same dates.
                self.assertEqual(boundary, res[i][b])



"""Tests for satip.eumetsat."""


def test_filename_to_datetime():
    """If there were a test here, there would also be a docstring here."""
    pass


def test_data_tailor():
    """If there were a test here, there would also be a docstring here."""
    pass



"""
Tests for the consumer app
"""
import datetime
import glob
import os
import tempfile

from click.testing import CliRunner
from freezegun import freeze_time

from satip.app import run

runner = CliRunner()


@freeze_time("2022-06-28 12:00:00")  # Date with RSS imagery
def test_save_to_netcdf():  # noqa 103
    user_key = os.environ.get("EUMETSAT_USER_KEY")
    user_secret = os.environ.get("EUMETSAT_USER_SECRET")
    with tempfile.TemporaryDirectory() as tmpdirname:
        runner.invoke(
            run,
            [
                "--api-key",
                user_key,
                "--api-secret",
                user_secret,
                "--save-dir",
                tmpdirname,
                "--use-rescaler",
                False,
                "--start-time",
                datetime.datetime.utcnow().isoformat(),
                "--maximum-n-datasets",
                1,
            ],
            catch_exceptions=False,
        )
        native_files = list(glob.glob(os.path.join(tmpdirname, "*.zarr.zip")))
        assert len(native_files) > 0


def test_save_to_netcdf_now():  # noqa 103
    user_key = os.environ.get("EUMETSAT_USER_KEY")
    user_secret = os.environ.get("EUMETSAT_USER_SECRET")
    with tempfile.TemporaryDirectory() as tmpdirname:
        runner.invoke(
            run,
            [
                "--api-key",
                user_key,
                "--api-secret",
                user_secret,
                "--save-dir",
                tmpdirname,
                "--use-rescaler",
                False,
                "--maximum-n-datasets",
                1,
            ],
            catch_exceptions=False,
        )
        native_files = list(glob.glob(os.path.join(tmpdirname, "*.zarr.zip")))
        assert len(native_files) > 0


def test_cleanup_now():  # noqa 103
    user_key = os.environ.get("EUMETSAT_USER_KEY")
    user_secret = os.environ.get("EUMETSAT_USER_SECRET")
    with tempfile.TemporaryDirectory() as tmpdirname:
        runner.invoke(
            run,
            [
                "--api-key",
                user_key,
                "--api-secret",
                user_secret,
                "--save-dir",
                tmpdirname,
                "--use-rescaler",
                False,
                "--cleanup",
                True,
                "--maximum-n-datasets",
                1,
            ],
            catch_exceptions=False,
        )
        native_files = list(glob.glob(os.path.join(tmpdirname, "*.zarr.zip")))
        assert len(native_files) == 0


@freeze_time("2022-06-22 12:00:00")  # Date with no RSS imagery
def test_save_datatailor_to_disk():  # noqa 103
    user_key = os.environ.get("EUMETSAT_USER_KEY")
    user_secret = os.environ.get("EUMETSAT_USER_SECRET")
    with tempfile.TemporaryDirectory() as tmpdirname:
        runner.invoke(
            run,
            [
                "--api-key",
                user_key,
                "--api-secret",
                user_secret,
                "--save-dir",
                tmpdirname,
                "--use-rescaler",
                False,
                "--start-time",
                datetime.datetime.utcnow().isoformat(),
                "--maximum-n-datasets",
                1,
            ],
            catch_exceptions=False,
        )
        native_files = list(glob.glob(os.path.join(tmpdirname, "*.zarr.zip")))
        assert len(native_files) > 0


@freeze_time("2022-06-28 12:00:00")  # Date with RSS imagery
def test_save_to_netcdf_rescaled():  # noqa 103
    user_key = os.environ.get("EUMETSAT_USER_KEY")
    user_secret = os.environ.get("EUMETSAT_USER_SECRET")
    with tempfile.TemporaryDirectory() as tmpdirname:
        runner.invoke(
            run,
            [
                "--api-key",
                user_key,
                "--api-secret",
                user_secret,
                "--save-dir",
                tmpdirname,
                "--use-rescaler",
                True,
                "--start-time",
                datetime.datetime.utcnow().isoformat(),
                "--maximum-n-datasets",
                1,
            ],
            catch_exceptions=False,
        )
        native_files = list(glob.glob(os.path.join(tmpdirname, "*.zarr.zip")))
        assert len(native_files) > 0


@freeze_time("2022-06-28 12:00:00")  # Use backup
def test_use_backup():  # noqa 103
    user_key = os.environ.get("EUMETSAT_USER_KEY")
    user_secret = os.environ.get("EUMETSAT_USER_SECRET")
    with tempfile.TemporaryDirectory() as tmpdirname:
        runner.invoke(
            run,
            [
                "--api-key",
                user_key,
                "--api-secret",
                user_secret,
                "--save-dir",
                tmpdirname,
                "--use-rescaler",
                False,
                "--start-time",
                datetime.datetime.utcnow().isoformat(),
                "--use-backup",
                True,
                "--maximum-n-datasets",
                1,
            ],
            catch_exceptions=False,
        )
        native_files = list(glob.glob(os.path.join(tmpdirname, "*.zarr.zip")))
        assert len(native_files) > 0



"""Thin wrapper around imagecodecs.JpegXl.
"""
from typing import Optional

import numpy as np
from imagecodecs.numcodecs import JpegXl
from numcodecs.registry import register_codec

# See the docs in `encode_nans` for an explanation of what these consts do.
# Also, please be aware that for sensical results, the values should follow the ordering:
# LOWER_BOUND_FOR_REAL_PIXELS > NAN_THRESHOLD > NAN_VALUE.
# Otherwise encoding and decoding back will lead to valid pixels being marked NaN.
LOWER_BOUND_FOR_REAL_PIXELS = 0.075
NAN_THRESHOLD = 0.05
NAN_VALUE = 0.025


class JpegXlFloatWithNaNs(JpegXl):
    """Thin wrapper around imagecodecs.JpegXl for floats with NaNs."""

    codec_id = "imagecodecs_jpegxl_float_with_nans"

    def __init__(
        self,
        lossless: Optional[bool] = None,
        distance: Optional[int] = None,
        level: Optional[int] = None,
        decodingspeed: Optional[float] = None,
        *args,
        **kwargs,
    ):
        """Initialise.

        This __init__ function is a simple hack to make the JpegXl compressor in the currently
        released version of imagecodecs (version 2021.11.20) look like the version in development.
        (We can't simply use the version in development because the imagecodecs author does not
        develop on GitHub. The imagecodecs authors just uses GitHub as one of several mechanisms
        to release imagecodecs.)

        See https://github.com/cgohlke/imagecodecs/issues/31#issuecomment-1026179413

        Args:
            lossless: Set to True to enable lossless compression.
            distance: Lowest settings are 0.00 or 0.01.  If 0.0 then also set lossless to True.
                To quote the cjxl man page:
                The preferred way to specify quality. It is specified in multiples of a
                just-noticeable difference. That is, -d 0 is mathematically lossless,
                -d 1 should be visually lossless, and higher distances yield denser and
                denser files with lower and lower fidelity.
            effort: To quote the cjxl man page:
                Controls the amount of effort that goes into producing an “optimal” file in
                terms of quality/size. That is to say, all other parameters being equal,
                a higher effort should yield a file that is at least as dense and possibly
                denser, and with at least as high and possibly higher quality.
                1 is fastest. 9 is slowest.
            level: DON'T SET THIS WITH THIS JpegXlFuture wrapper!
                In imagecodecs version 2021.11.20, level is mapped (incorrectly) to the decoding
                speed tier. Minimum is 0 (highest quality), and maximum is 4 (lowest quality).
                Default is 0.
            decodingspeed: DON'T SET THIS WITH THIS JpegXlFuture wrapper!
        """
        assert decodingspeed is None
        lossless = False
        if lossless is not None:
            if lossless:
                assert (
                    level is None
                )  # level must be None to enable lossless in imagecodecs 2021.11.20.
                assert distance is None or distance == 0
            else:
                # Enable lossy compression.
                # level must be set to 0, 1, 2, 3, or 4 to enable lossy
                # compression in imagecodecs 2021.11.20.
                level = 0
        super().__init__(level=level, distance=distance, *args, **kwargs)

    def encode(self, buf: np.ndarray) -> None:
        """Encode `buf` with JPEG-XL.

        Under the hood, NaNs will be encoded as NAN_VALUE, and all "real"
        values will be in the range [LOWER_BOUND_FOR_REAL_PIXELS, 1]. But
        this is all taken care of by encode and decode.

        Args:
            buf: The input array to compress as JPEG-XL.
                Expects buf to be of shape (n_timesteps, y, x, n_channels).
                n_timesteps and n_channels must == 1.
                buf can be float16 or float32.
                All values must be in the range [0, 1].
                Use as much of the range [0, 1] as possible. 0 is black and 1 is white.
                If all the information is squished into, say, the range [0, 0.1] then
                JPEG-XL will interpret the image as very dark, and will agressively
                compress the data because JPEG-XL assumes that human viewers do not
                notice if detail is lost in the shadows.
        """
        assert buf.dtype in (
            np.float16,
            np.float32,
        ), f"dtype must be float16 or float32, not {buf.dtype}"

        # Even if the original DataArray doesn't have NaNs,
        # when Zarr saves chunks at the edges of the image, the image data for that chunk
        # might be smaller than the chunk. In that case, `buf` will be the same shape
        # as the nominal chunk size, but will include NaNs. Those NaNs must be encoded
        # with a floating point value in the range [0, 1] because JPEG-XL cannot
        # handle NaN values.
        buf = encode_nans(buf)

        # Sanity checks.
        assert np.all(np.isfinite(buf))
        assert np.amin(buf) >= 0
        assert np.amax(buf) <= 1

        # In the future, JPEG-XL will be able to encode multiple images per file.
        # But, for now, it can only compress one image at a time. See:
        # https://github.com/cgohlke/imagecodecs/issues/32
        n_timesteps = buf.shape[0]
        n_channels = buf.shape[-1]
        assert n_timesteps == 1
        assert n_channels == 1

        return super().encode(buf)

    def decode(self, buf, out=None) -> np.ndarray:
        """Decode JPEG-XL `buf`.

        Reconstruct the NaNs encoded by encode_nans.
        """
        out = super().decode(buf, out)
        out = decode_nans(out)
        return out


def encode_nans(data: np.ndarray) -> np.ndarray:
    """Encode NaNs as the value NAN_VALUE.

    Encode all other values in the range [LOWER_BOUND_FOR_REAL_PIXELS, 1].

    JPEG-XL does not understand "NaN" values. JPEG-XL only understands floating
    point values in the range [0, 1]. So we must encode NaN values
    as real values in the range [0, 1].

    After lossy JPEG-XL compression, there is slight "ringing" around the edges
    of regions with filled with a constant number. In experiments, when NAN_VALUE = 0.025,
    it appears that the values at the inner edges of a "NaN region" vary in the range
    [0.0227, 0.0280]. But, to be safe, we use a nice wide margin: We don't set
    the value of "NaNs" to be 0.00 because the ringing would cause the values
    to drop below zero, which is illegal for JPEG-XL images.

    After decompression, reconstruct regions of NaNs using "image < NAN_THRESHOLD"
    to find NaNs.

    See this comment for more info:
    https://github.com/openclimatefix/Satip/issues/67#issuecomment-1036456502

    Args:
        data: The input data. All values must already
            be in the range [0, 1]. The original data is modified in place.

    Returns:
        The returned array. "Real" values will be shifted to
            the range [LOWER_BOUND_FOR_REAL_PIXELS, 1].
            NaNs will be encoded as NAN_VALUE.
    """
    assert issubclass(
        data.dtype.type, np.floating
    ), f"dataarray.dtype must be floating point not {data.dtype}!"

    # Shift all the "real" values up to the range [0.075, 1]
    data *= 1 - LOWER_BOUND_FOR_REAL_PIXELS
    # Now [0, 1-0.075]
    data += LOWER_BOUND_FOR_REAL_PIXELS
    # Now [0.075, 1.]
    data = np.nan_to_num(data, nan=NAN_VALUE).clip(min=0, max=1)
    return data


def decode_nans(data: np.ndarray) -> np.ndarray:
    """Reconstruct the NaNs encoded by encode_nans."""
    assert np.all(np.isfinite(data))
    assert issubclass(data.dtype.type, np.floating)
    data[data <= NAN_THRESHOLD] = np.NaN
    # [0.075, 1]
    data -= LOWER_BOUND_FOR_REAL_PIXELS
    # [0, 1-0.075]
    data /= 1 - LOWER_BOUND_FOR_REAL_PIXELS
    # [0, 1]
    return data.clip(min=0, max=1)


register_codec(JpegXlFloatWithNaNs)
